-------------------------------------------------------------------------------
--	SimpleSystem CPU [MEMORY I/F]											 --
--	Copyright(C)2012 t.hara													 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/08/15	t.hara	first release										 --
-------------------------------------------------------------------------------

library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_signed.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_cpu_memory is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;
		sreset			: in	std_logic;
		-- request from state
		fetch_address	: in	std_logic_vector( 31 downto 0 );
		fetch_oe		: in	std_logic;
		fetch_q			: out	std_logic_vector( 15 downto 0 );
		fetch_q_en		: out	std_logic;
		-- request from decoder
		dec_address		: in	std_logic_vector( 31 downto 0 );
		dec_d			: in	std_logic_vector( 31 downto 0 );
		dec_width		: in	std_logic_vector(  1 downto 0 );	-- "00": 1byte, "01": 2byte, "1-": 4byte
		dec_we			: in	std_logic;
		dec_oe			: in	std_logic;
		dec_q			: out	std_logic_vector( 31 downto 0 );
		dec_q_en		: out	std_logic;
		-- request to 16bit bus interface
		cpu_address		: out	std_logic_vector( 30 downto 0 );
		cpu_d			: out	std_logic_vector( 15 downto 0 );
		cpu_we0			: out	std_logic;							-- lower byte
		cpu_we1			: out	std_logic;							-- upper byte
		cpu_oe			: out	std_logic;
		cpu_q			: in	std_logic_vector( 15 downto 0 );
		cpu_q_en		: in	std_logic
	);
end simplesystem_cpu_memory;

architecture rtl of simplesystem_cpu_memory is
	signal ff_cpu_address		: std_logic_vector( 30 downto 0 );
	signal ff_cpu_d				: std_logic_vector( 15 downto 0 );
	signal ff_cpu_we0			: std_logic;
	signal ff_cpu_we1			: std_logic;
	signal ff_cpu_oe			: std_logic;
	signal ff_cpu_qh			: std_logic_vector( 15 downto 0 );
	signal ff_cpu_ql			: std_logic_vector( 15 downto 0 );
	signal ff_cpu_fetch			: std_logic_vector( 15 downto 0 );
	signal ff_cpu_q_en			: std_logic;
	signal ff_state				: std_logic;
	signal ff_delay				: std_logic;
	signal ff_byteswap			: std_logic;
begin

	-- out assignment
	fetch_q			<= ff_cpu_fetch;
	fetch_q_en		<= ff_cpu_q_en and (not ff_delay);
	dec_q			<= ff_cpu_qh & ff_cpu_ql;
	dec_q_en		<= ff_cpu_q_en and (not ff_delay);
	cpu_address		<= ff_cpu_address;
	cpu_d			<= ff_cpu_d;
	cpu_we0			<= ff_cpu_we0;
	cpu_we1			<= ff_cpu_we1;
	cpu_oe			<= ff_cpu_oe;

	-- state
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_state <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_state <= '0';
			elsif( (dec_we or dec_oe) = '1' and dec_width(1) = '1' )then
				ff_state <= '1';
			else
				ff_state <= '0';
			end if;
		end if;
	end process;

	-- state delay for receive
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_delay <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_delay <= '0';
			elsif( ff_state = '1' )then
				ff_delay <= '1';
			elsif( ff_cpu_q_en = '1' )then
				ff_delay <= '0';
			end if;
		end if;
	end process;

	-- latch
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_cpu_address	<= (others => '0');
			ff_cpu_d		<= (others => '0');
			ff_cpu_we0		<= '0';
			ff_cpu_we1		<= '0';
			ff_cpu_oe		<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_cpu_address	<= (others => '0');
				ff_cpu_d		<= (others => '0');
				ff_cpu_we0		<= '0';
				ff_cpu_we1		<= '0';
				ff_cpu_oe		<= '0';
			elsif( fetch_oe = '1' )then
				ff_cpu_address	<= fetch_address(31 downto 1);
				ff_cpu_d		<= (others => '0');
				ff_cpu_we0		<= '0';
				ff_cpu_we1		<= '0';
				ff_cpu_oe		<= '1';
			elsif( dec_we = '1' and dec_width = "00" )then
				ff_cpu_address	<= dec_address(31 downto 1);
				ff_cpu_d		<= dec_d( 7 downto 0 ) & dec_d( 7 downto 0 );
				ff_cpu_we0		<= not dec_address(0);
				ff_cpu_we1		<= dec_address(0);
				ff_cpu_oe		<= '0';
			elsif( dec_we = '1' and dec_width = "01" )then
				ff_cpu_address	<= dec_address(31 downto 1);
				ff_cpu_d		<= dec_d( 15 downto 0 );
				ff_cpu_we0		<= '1';
				ff_cpu_we1		<= '1';
				ff_cpu_oe		<= '0';
			elsif( dec_we = '1' and dec_width(1) = '1' )then
				ff_cpu_address	<= dec_address(31 downto 1);
				ff_cpu_d		<= dec_d( 15 downto  0 );
				ff_cpu_we0		<= '1';
				ff_cpu_we1		<= '1';
				ff_cpu_oe		<= '0';
			elsif( dec_oe = '1' and dec_width(1) = '0' )then
				ff_cpu_address	<= dec_address(31 downto 1);
				ff_cpu_d		<= (others => '0');
				ff_cpu_we0		<= '0';
				ff_cpu_we1		<= '0';
				ff_cpu_oe		<= '1';
			elsif( dec_oe = '1' and dec_width(1) = '1' )then
				ff_cpu_address	<= dec_address(31 downto 1);
				ff_cpu_d		<= (others => '0');
				ff_cpu_we0		<= '0';
				ff_cpu_we1		<= '0';
				ff_cpu_oe		<= '1';
			elsif( ff_state = '1' )then
				ff_cpu_address	<= ff_cpu_address + 1;
				ff_cpu_d		<= dec_d( 31 downto 16 );
			elsif( ff_state = '0' )then
				ff_cpu_we0		<= '0';
				ff_cpu_we1		<= '0';
				ff_cpu_oe		<= '0';
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_cpu_q_en <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_cpu_q_en <= '0';
			else
				ff_cpu_q_en <= cpu_q_en;
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_cpu_ql		<= (others => '0');
			ff_cpu_fetch	<= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_cpu_ql		<= (others => '0');
				ff_cpu_fetch	<= (others => '0');
			elsif( (cpu_q_en and not ff_cpu_q_en) = '1' )then
				if( ff_byteswap = '0' )then
					ff_cpu_ql		<= cpu_q;
				else
					ff_cpu_ql		<= cpu_q( 15 downto 8 ) & cpu_q( 15 downto 8 );
				end if;
				ff_cpu_fetch	<= cpu_q;
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_cpu_qh <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_cpu_qh <= (others => '0');
			elsif( (cpu_q_en and ff_cpu_q_en) = '1' )then
				if( ff_byteswap = '0' )then
					ff_cpu_qh <= cpu_q;
				else
					ff_cpu_qh <= cpu_q( 15 downto 8 ) & cpu_q( 15 downto 8 );
				end if;
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_byteswap <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_byteswap <= '0';
			elsif( dec_oe = '1' )then
				if( dec_width = "00" )then
					ff_byteswap <= dec_address(0);
				else
					ff_byteswap <= '0';
				end if;
			end if;
		end if;
	end process;
end rtl;
