-------------------------------------------------------------------------------
--	SimpleSystem CPU [REGISTER FILE]										 --
--	Copyright(C)2012 t.hara													 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/08/15	t.hara	first release										 --
-------------------------------------------------------------------------------

library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_signed.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_cpu_register is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;
		sreset			: in	std_logic;
		-- update target
		dst_d			: in	std_logic_vector( 31 downto 0 );
		dst_sel			: in	std_logic_vector(  2 downto 0 );
		dst_we			: in	std_logic;
		irqvec_d		: in	std_logic_vector( 31 downto 0 );
		irqvec_we		: in	std_logic;
		irqen_d			: in	std_logic;
		irqen_we		: in	std_logic;
		dec_mem_width	: in	std_logic_vector(  1 downto 0 );
		dec_mem_oe		: in	std_logic;
		dec_mem_q		: in	std_logic_vector( 31 downto 0 );
		dec_mem_q_en	: in	std_logic;
		-- interrupt request
		irq				: in	std_logic;
		state0			: in	std_logic;
		state4			: in	std_logic;
		set32			: in	std_logic;
		-- output
		r0				: out	std_logic_vector( 31 downto 0 );
		r1				: out	std_logic_vector( 31 downto 0 );
		r2				: out	std_logic_vector( 31 downto 0 );
		r3				: out	std_logic_vector( 31 downto 0 );
		r4				: out	std_logic_vector( 31 downto 0 );
		r5				: out	std_logic_vector( 31 downto 0 );
		r6				: out	std_logic_vector( 31 downto 0 );
		r7				: out	std_logic_vector( 31 downto 0 );
		irqret			: out	std_logic_vector( 31 downto 0 )
	);
end simplesystem_cpu_register;

architecture rtl of simplesystem_cpu_register is
	signal ff_r0			: std_logic_vector( 31 downto 0 );
	signal ff_r1			: std_logic_vector( 31 downto 0 );
	signal ff_r2			: std_logic_vector( 31 downto 0 );
	signal ff_r3			: std_logic_vector( 31 downto 0 );
	signal ff_r4			: std_logic_vector( 31 downto 0 );
	signal ff_r5			: std_logic_vector( 31 downto 0 );
	signal ff_r6			: std_logic_vector( 31 downto 0 );
	signal ff_r7			: std_logic_vector( 31 downto 0 );
	signal ff_irqvec		: std_logic_vector( 31 downto 0 );
	signal ff_irqret		: std_logic_vector( 31 downto 0 );
	signal ff_irqen			: std_logic;
	signal ff_mem_width		: std_logic_vector(  1 downto 0 );
	signal w_irq_happen		: std_logic;
	signal w_dst_d			: std_logic_vector( 31 downto 0 );
	signal w_dst_sel		: std_logic_vector(  2 downto 0 );
	constant c_zero			: std_logic_vector( 31 downto 8 ) := (others => '0');
begin

	-- output assignment
	r0		<= ff_r0;
	r1		<= ff_r1;
	r2		<= ff_r2;
	r3		<= ff_r3;
	r4		<= ff_r4;
	r5		<= ff_r5;
	r6		<= ff_r6;
	r7		<= ff_irqvec when( w_irq_happen = '1' )else ff_r7;
	irqret	<= ff_irqret;

	-- interrupt
	w_irq_happen	<=	irq and state0 and ff_irqen;

	-- write data
	w_dst_d			<=	dst_d												when( dec_mem_q_en = '0' )else
						c_zero( 31 downto  8 ) & dec_mem_q(  7 downto 0 )	when( ff_mem_width = "00" )else
						c_zero( 31 downto 16 ) & dec_mem_q( 15 downto 0 )	when( ff_mem_width = "01" )else
						dec_mem_q;

	w_dst_sel		<=	dst_sel;

	-- register blocks
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_r0 <= (others => '0');
			ff_r1 <= (others => '0');
			ff_r2 <= (others => '0');
			ff_r3 <= (others => '0');
			ff_r4 <= (others => '0');
			ff_r5 <= (others => '0');
			ff_r6 <= (others => '0');
			ff_r7 <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_r0 <= (others => '0');
				ff_r1 <= (others => '0');
				ff_r2 <= (others => '0');
				ff_r3 <= (others => '0');
				ff_r4 <= (others => '0');
				ff_r5 <= (others => '0');
				ff_r6 <= (others => '0');
				ff_r7 <= (others => '0');
			elsif( w_irq_happen = '1' )then
				ff_r7 <= ff_irqvec + 2;
			elsif( state0 = '1' )then
				ff_r7 <= ff_r7 + 2;
			elsif( set32 = '1' )then
				ff_r7 <= ff_r7 + 4;
			elsif( ((state4 and dec_mem_q_en) or dst_we) = '1' )then
				case w_dst_sel is
					when "000"  => ff_r0 <= w_dst_d;
					when "001"  => ff_r1 <= w_dst_d;
					when "010"  => ff_r2 <= w_dst_d;
					when "011"  => ff_r3 <= w_dst_d;
					when "100"  => ff_r4 <= w_dst_d;
					when "101"  => ff_r5 <= w_dst_d;
					when "110"  => ff_r6 <= w_dst_d;
					when "111"  => ff_r7 <= w_dst_d;
					when others => -- hold
				end case;
			end if;
		end if;
	end process;

	-- interrupt vector
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_irqvec <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_irqvec <= (others => '0');
			elsif( irqvec_we = '1' )then
				ff_irqvec <= irqvec_d;
			end if;
		end if;
	end process;

	-- interrupt mask
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_irqen <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_irqen <= '0';
			elsif( w_irq_happen = '1' )then
				ff_irqen <= '0';
			elsif( irqen_we = '1' )then
				ff_irqen <= irqen_d;
			end if;
		end if;
	end process;

	-- interrupt return address
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_irqret <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_irqret <= (others => '0');
			elsif( w_irq_happen = '1' )then
				ff_irqret <= ff_r7;
			end if;
		end if;
	end process;

	-- memory read width
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mem_width <= "00";
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_mem_width <= "00";
			elsif( dec_mem_oe = '1' )then
				ff_mem_width <= dec_mem_width;
			end if;
		end if;
	end process;
end rtl;
