-------------------------------------------------------------------------------
--	SimpleSystem CPU [TOP ENTITY]											 --
--	Copyright(C)2012 t.hara													 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/08/16	t.hara	first release										 --
-------------------------------------------------------------------------------

library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_signed.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_cpu is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;
		sreset			: in	std_logic;
		-- interrupt request
		irq				: in	std_logic;
		-- bus interface
		cpu_address		: out	std_logic_vector( 30 downto 0 );
		cpu_d			: out	std_logic_vector( 15 downto 0 );
		cpu_we0			: out	std_logic;							-- lower byte
		cpu_we1			: out	std_logic;							-- upper byte
		cpu_oe			: out	std_logic;
		cpu_q			: in	std_logic_vector( 15 downto 0 );
		cpu_q_en		: in	std_logic
	);
end simplesystem_cpu;

architecture rtl of simplesystem_cpu is
	component simplesystem_cpu_state
		port(
			reset			: in	std_logic;
			clk				: in	std_logic;
			sreset			: in	std_logic;
			-- state notify
			state0			: out	std_logic;
			state2			: out	std_logic;
			state4			: out	std_logic;
			-- update signal from decoder
			dec_mem_we		: in	std_logic;
			dec_mem_oe		: in	std_logic;
			-- destination
			mem_oe			: out	std_logic;
			mem_ack			: in	std_logic
		);
	end component;

	component simplesystem_cpu_memory
		port(
			reset			: in	std_logic;
			clk				: in	std_logic;
			sreset			: in	std_logic;
			-- request from state
			fetch_address	: in	std_logic_vector( 31 downto 0 );
			fetch_oe		: in	std_logic;
			fetch_q			: out	std_logic_vector( 15 downto 0 );
			fetch_q_en		: out	std_logic;
			-- request from decoder
			dec_address		: in	std_logic_vector( 31 downto 0 );
			dec_d			: in	std_logic_vector( 31 downto 0 );
			dec_width		: in	std_logic_vector(  1 downto 0 );	-- "00": 1byte, "01": 2byte, "1-": 4byte
			dec_we			: in	std_logic;
			dec_oe			: in	std_logic;
			dec_q			: out	std_logic_vector( 31 downto 0 );
			dec_q_en		: out	std_logic;
			-- request to 16bit bus interface
			cpu_address		: out	std_logic_vector( 30 downto 0 );
			cpu_d			: out	std_logic_vector( 15 downto 0 );
			cpu_we0			: out	std_logic;							-- lower byte
			cpu_we1			: out	std_logic;							-- upper byte
			cpu_oe			: out	std_logic;
			cpu_q			: in	std_logic_vector( 15 downto 0 );
			cpu_q_en		: in	std_logic
		);
	end component;

	component simplesystem_cpu_decoder
		port(
			reset			: in	std_logic;
			clk				: in	std_logic;
			sreset			: in	std_logic;
			-- code
			opecode			: in	std_logic_vector( 15 downto 0 );
			state2			: in	std_logic;
			-- register file
			r0				: in	std_logic_vector( 31 downto 0 );
			r1				: in	std_logic_vector( 31 downto 0 );
			r2				: in	std_logic_vector( 31 downto 0 );
			r3				: in	std_logic_vector( 31 downto 0 );
			r4				: in	std_logic_vector( 31 downto 0 );
			r5				: in	std_logic_vector( 31 downto 0 );
			r6				: in	std_logic_vector( 31 downto 0 );
			r7				: in	std_logic_vector( 31 downto 0 );
			irqret			: in	std_logic_vector( 31 downto 0 );
			-- destination
			set32			: out	std_logic;
			dst_d			: out	std_logic_vector( 31 downto 0 );
			dst_sel			: out	std_logic_vector(  2 downto 0 );
			dst_we			: out	std_logic;
			irqvec_d		: out	std_logic_vector( 31 downto 0 );
			irqvec_we		: out	std_logic;
			irqen_d			: out	std_logic;
			irqen_we		: out	std_logic;
			mem_address		: out	std_logic_vector( 31 downto 0 );
			mem_width		: out	std_logic_vector(  1 downto 0 );
			mem_d			: out	std_logic_vector( 31 downto 0 );
			mem_we			: out	std_logic;
			mem_oe			: out	std_logic;
			state_we		: out	std_logic;
			state_oe		: out	std_logic
		);
	end component;

	component simplesystem_cpu_register
		port(
			reset			: in	std_logic;
			clk				: in	std_logic;
			sreset			: in	std_logic;
			-- update target
			dst_d			: in	std_logic_vector( 31 downto 0 );
			dst_sel			: in	std_logic_vector(  2 downto 0 );
			dst_we			: in	std_logic;
			irqvec_d		: in	std_logic_vector( 31 downto 0 );
			irqvec_we		: in	std_logic;
			irqen_d			: in	std_logic;
			irqen_we		: in	std_logic;
			dec_mem_width	: in	std_logic_vector(  1 downto 0 );
			dec_mem_oe		: in	std_logic;
			dec_mem_q		: in	std_logic_vector( 31 downto 0 );
			dec_mem_q_en	: in	std_logic;
			-- interrupt request
			irq				: in	std_logic;
			state0			: in	std_logic;
			state4			: in	std_logic;
			set32			: in	std_logic;
			-- output
			r0				: out	std_logic_vector( 31 downto 0 );
			r1				: out	std_logic_vector( 31 downto 0 );
			r2				: out	std_logic_vector( 31 downto 0 );
			r3				: out	std_logic_vector( 31 downto 0 );
			r4				: out	std_logic_vector( 31 downto 0 );
			r5				: out	std_logic_vector( 31 downto 0 );
			r6				: out	std_logic_vector( 31 downto 0 );
			r7				: out	std_logic_vector( 31 downto 0 );
			irqret			: out	std_logic_vector( 31 downto 0 )
		);
	end component;

	signal state0			: std_logic;
	signal state2			: std_logic;
	signal state4			: std_logic;
	signal set32			: std_logic;
	signal r0				: std_logic_vector( 31 downto 0 );
	signal r1				: std_logic_vector( 31 downto 0 );
	signal r2				: std_logic_vector( 31 downto 0 );
	signal r3				: std_logic_vector( 31 downto 0 );
	signal r4				: std_logic_vector( 31 downto 0 );
	signal r5				: std_logic_vector( 31 downto 0 );
	signal r6				: std_logic_vector( 31 downto 0 );
	signal r7				: std_logic_vector( 31 downto 0 );
	signal irqret			: std_logic_vector( 31 downto 0 );
	signal fetch_oe			: std_logic;
	signal fetch_q			: std_logic_vector( 15 downto 0 );
	signal fetch_q_en		: std_logic;
	signal dst_d			: std_logic_vector( 31 downto 0 );
	signal dst_sel			: std_logic_vector(  2 downto 0 );
	signal dst_we			: std_logic;
	signal irqvec_d			: std_logic_vector( 31 downto 0 );
	signal irqvec_we		: std_logic;
	signal irqen_d			: std_logic;
	signal irqen_we			: std_logic;
	signal dec_mem_address	: std_logic_vector( 31 downto 0 );
	signal dec_mem_width	: std_logic_vector(  1 downto 0 );
	signal dec_mem_d		: std_logic_vector( 31 downto 0 );
	signal dec_mem_we		: std_logic;
	signal dec_mem_oe		: std_logic;
	signal dec_mem_q		: std_logic_vector( 31 downto 0 );
	signal dec_mem_q_en		: std_logic;
	signal state_we			: std_logic;
	signal state_oe			: std_logic;
begin

	u_state: simplesystem_cpu_state
	port map(
		reset			=> reset			,
		clk				=> clk				,
		sreset			=> sreset			,
		state0			=> state0			,
		state2			=> state2			,
		state4			=> state4			,
		dec_mem_we		=> state_we			,
		dec_mem_oe		=> state_oe			,
		mem_oe			=> fetch_oe			,
		mem_ack			=> fetch_q_en		
	);

	u_memory: simplesystem_cpu_memory
	port map(
		reset			=> reset			,
		clk				=> clk				,
		sreset			=> sreset			,
		fetch_address	=> r7				,
		fetch_oe		=> fetch_oe			,
		fetch_q			=> fetch_q			,
		fetch_q_en		=> fetch_q_en		,
		dec_address		=> dec_mem_address	,
		dec_d			=> dec_mem_d		,
		dec_width		=> dec_mem_width	,
		dec_we			=> dec_mem_we		,
		dec_oe			=> dec_mem_oe		,
		dec_q			=> dec_mem_q		,
		dec_q_en		=> dec_mem_q_en		,
		cpu_address		=> cpu_address		,
		cpu_d			=> cpu_d			,
		cpu_we0			=> cpu_we0			,
		cpu_we1			=> cpu_we1			,
		cpu_oe			=> cpu_oe			,
		cpu_q			=> cpu_q			,
		cpu_q_en		=> cpu_q_en			
	);

	u_decoder: simplesystem_cpu_decoder
	port map(
		reset			=> reset			,
		clk				=> clk				,
		sreset			=> sreset			,
		opecode			=> fetch_q			,
		state2			=> state2			,
		r0				=> r0				,
		r1				=> r1				,
		r2				=> r2				,
		r3				=> r3				,
		r4				=> r4				,
		r5				=> r5				,
		r6				=> r6				,
		r7				=> r7				,
		irqret			=> irqret			,
		set32			=> set32			,
		dst_d			=> dst_d			,
		dst_sel			=> dst_sel			,
		dst_we			=> dst_we			,
		irqvec_d		=> irqvec_d			,
		irqvec_we		=> irqvec_we		,
		irqen_d			=> irqen_d			,
		irqen_we		=> irqen_we			,
		mem_address		=> dec_mem_address	,
		mem_width		=> dec_mem_width	,
		mem_d			=> dec_mem_d		,
		mem_we			=> dec_mem_we		,
		mem_oe			=> dec_mem_oe		,
		state_we		=> state_we			,
		state_oe		=> state_oe			
	);

	u_register: simplesystem_cpu_register
	port map(
		reset			=> reset			,
		clk				=> clk				,
		sreset			=> sreset			,
		dst_d			=> dst_d			,
		dst_sel			=> dst_sel			,
		dst_we			=> dst_we			,
		irqvec_d		=> irqvec_d			,
		irqvec_we		=> irqvec_we		,
		irqen_d			=> irqen_d			,
		irqen_we		=> irqen_we			,
		dec_mem_width	=> dec_mem_width	,
		dec_mem_oe		=> dec_mem_oe		,
		dec_mem_q		=> dec_mem_q		,
		dec_mem_q_en	=> dec_mem_q_en		,
		irq				=> irq				,
		state0			=> state0			,
		state4			=> state4			,
		set32			=> set32			,
		r0				=> r0				,
		r1				=> r1				,
		r2				=> r2				,
		r3				=> r3				,
		r4				=> r4				,
		r5				=> r5				,
		r6				=> r6				,
		r7				=> r7				,
		irqret			=> irqret			
	);
end rtl;
