-------------------------------------------------------------------------------
--	SimpleSystem SD/MMC card controller										 --
--	Copyright(C)2012 t.hara													 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/09/22	t.hara	first release										 --
-------------------------------------------------------------------------------
--	Address																	 --
--		0	w	chip select (0: disable, 1: enable)							 --
--		1	w	data send													 --
--		1	r	data receive												 --
-------------------------------------------------------------------------------
library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_unsigned.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_sdcard is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;	-- 50MHz
		sreset			: in	std_logic;
		sdc_a			: in	std_logic;
		sdc_d			: in	std_logic_vector( 15 downto 0 );
		sdc_oe			: in	std_logic;
		sdc_we0			: in	std_logic;
		sdc_we1			: in	std_logic;
		sdc_q			: out	std_logic_vector( 15 downto 0 );
		sdc_q_en		: out	std_logic;
		-- to SD/MMC card (SPI mode)
		sd_clk			: out	std_logic;	-- mmc_clk
		sd_cmd			: in	std_logic;	-- mmc_di
		sd_dat3			: out	std_logic;	-- mmc_cs
		sd_dat			: out	std_logic	-- mmc_do
	);
end simplesystem_sdcard;

architecture rtl of simplesystem_sdcard is
	alias  mmc_clk			: std_logic is sd_clk;
	alias  mmc_di			: std_logic is sd_cmd;
	alias  mmc_cs			: std_logic is sd_dat3;
	alias  mmc_do			: std_logic is sd_dat;

	signal ff_q				: std_logic_vector(  7 downto 0 );
	signal ff_q_en			: std_logic;
	signal ff_bits			: std_logic_vector(  7 downto 0 );
	signal ff_mmc_clk		: std_logic;
	signal ff_mmc_clk_div	: std_logic;
	signal ff_mmc_cs		: std_logic;
	signal ff_mmc_do		: std_logic;
	signal ff_mmc_di		: std_logic;
	signal ff_active		: std_logic;
	signal ff_remain_bit	: std_logic_vector(  2 downto 0 );
	signal ff_direction		: std_logic;
	signal ff_direction_d1	: std_logic;
	signal w_disactive		: std_logic;
	signal w_rise_edge		: std_logic;
	signal w_fall_edge		: std_logic;
	signal w_receive_bit	: std_logic;
begin

	-- output assignment (SD/MMC card)
	mmc_clk		<= ff_mmc_clk;
	mmc_cs		<= ff_mmc_cs;
	mmc_do		<= ff_mmc_do when( ff_direction_d1 = '1' )else 'Z';

	-- output assignment
	sdc_q		<= "00000000" & ff_q;
	sdc_q_en	<= ff_q_en;

	-- SD clock generator (50MHz --> 12.5MHz)
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mmc_clk_div <= '0';
		elsif( clk'event and clk = '1' )then
			if( (sreset or (not ff_active)) = '1' )then
				ff_mmc_clk_div <= '0';
			else
				ff_mmc_clk_div <= not ff_mmc_clk_div;
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mmc_clk <= '0';
		elsif( clk'event and clk = '1' )then
			if( (sreset or (not ff_active)) = '1' )then
				ff_mmc_clk <= '0';
			elsif( ff_mmc_clk_div = '1' )then
				ff_mmc_clk <= not ff_mmc_clk;
			end if;
		end if;
	end process;

	w_rise_edge <= ff_mmc_clk_div and (not ff_mmc_clk);
	w_fall_edge <= ff_mmc_clk_div and ff_mmc_clk;

	-- chip select
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mmc_cs <= '1';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_mmc_cs <= '1';
			elsif( sdc_a = '0' and sdc_we0 = '1' )then
				ff_mmc_cs <= not sdc_d(0);
			end if;
		end if;
	end process;

	-- active
	w_disactive <= '1' when( ff_remain_bit = "000" )else '0';

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_active <= '0';
			ff_direction <= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_active <= '0';
				ff_direction <= '0';
			elsif( sdc_a = '1' and (sdc_oe or sdc_we0) = '1' )then
				ff_active <= '1';
				ff_direction <= sdc_we0;
			elsif( (w_disactive and w_fall_edge) = '1' )then
				ff_active <= '0';
				ff_direction <= '0';
			end if;
		end if;
	end process;

	-- send/receive bit count
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_remain_bit <= "000";
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_remain_bit <= "000";
			elsif( sdc_a = '1' and (sdc_oe or sdc_we0) = '1' )then
				ff_remain_bit <= "111";
			elsif( w_disactive = '1' )then
				-- hold
			elsif( w_fall_edge = '1' )then
				ff_remain_bit <= ff_remain_bit - 1;
			end if;
		end if;
	end process;

	-- send/receive bits
	w_receive_bit <= ff_mmc_di when( ff_direction = '0' )else '0';

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_bits <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_bits <= (others => '0');
			elsif( sdc_a = '1' and sdc_we0 = '1' )then
				ff_bits <= sdc_d(  7 downto 0 );
			elsif( (ff_active and w_fall_edge) = '1' )then
				ff_bits <= ff_bits( 6 downto 0 ) & w_receive_bit;
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mmc_do <= '0';
			ff_mmc_di <= '0';
			ff_direction_d1 <= '0';
		elsif( clk'event and clk = '1' )then
			ff_mmc_do <= ff_bits(7);
			ff_direction_d1 <= ff_direction;
			ff_mmc_di <= mmc_di;
		end if;
	end process;

	-- receive bits
	process( reset, clk )
	begin
		if( reset = '1' )then
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_q	<= (others => '0');
				ff_q_en	<= '0';
			elsif( sdc_a = '0' and (sdc_oe or sdc_we0) = '1' )then
				ff_q	<= sdc_d( 7 downto 0 );
				ff_q_en	<= '1';
			elsif( (w_disactive and w_fall_edge and ff_active) = '1' )then
				ff_q	<= ff_bits;
				ff_q_en	<= '1';
			else
				ff_q	<= (others => '0');
				ff_q_en	<= '0';
			end if;
		end if;
	end process;
end rtl;
