-------------------------------------------------------------------------------
--	SRAM for Altera DE1 (IS61LV25616-10TL)									 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/08/20	t.hara														 --
-------------------------------------------------------------------------------

library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_signed.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_sram is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;
		sreset			: in	std_logic;
		ram_a			: in	std_logic_vector( 17 downto 0 );
		ram_d			: in	std_logic_vector( 15 downto 0 );
		ram_oe			: in	std_logic;
		ram_we0			: in	std_logic;
		ram_we1			: in	std_logic;
		ram_q			: out	std_logic_vector( 15 downto 0 );
		ram_q_en		: out	std_logic;
		-- to SRAM
		sram_addr		: out	std_logic_vector( 17 downto 0 );
		sram_dq			: inout	std_logic_vector( 15 downto 0 );
		sram_we_n		: out	std_logic;
		sram_oe_n		: out	std_logic;
		sram_ub_n		: out	std_logic;
		sram_lb_n		: out	std_logic;
		sram_ce_n		: out	std_logic
	);
end simplesystem_sram;

architecture rtl of simplesystem_sram is

	signal ff_we_n		: std_logic;
	signal ff_oe_n		: std_logic;
	signal ff_ub_n		: std_logic;
	signal ff_lb_n		: std_logic;
	signal ff_ce_n		: std_logic;

	signal ff_addr		: std_logic_vector( 17 downto 0 );
	signal ff_q_en		: std_logic;
	signal ff_d			: std_logic_vector( 15 downto 0 );
	signal ff_q			: std_logic_vector( 15 downto 0 );
begin

	-- output assignment for SRAM
	sram_addr	<=	ff_addr;
	sram_dq		<=	ff_d		when( ff_we_n = '0' )else
					(others => 'Z');
	sram_we_n	<=	ff_we_n;
	sram_oe_n	<=	ff_oe_n;
	sram_ub_n	<=	ff_ub_n;
	sram_lb_n	<=	ff_lb_n;
	sram_ce_n	<=	ff_ce_n;

	-- output assignment
	ram_q		<=	ff_q;
	ram_q_en	<=	ff_q_en;

	-- chip enable
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_we_n	<= '1';
			ff_oe_n	<= '1';
			ff_lb_n	<= '1';
			ff_ub_n	<= '1';
			ff_ce_n	<= '1';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_we_n	<= '1';
				ff_oe_n	<= '1';
				ff_lb_n	<= '1';
				ff_ub_n	<= '1';
				ff_ce_n	<= '1';
			else
				ff_we_n	<= not( ram_we0 or ram_we1 );
				ff_oe_n	<= not  ram_oe;
				ff_lb_n	<= not( ram_oe or ram_we0 );
				ff_ub_n	<= not( ram_oe or ram_we1 );
				ff_ce_n	<= not( ram_oe or ram_we0 or ram_we1 );
			end if;
		end if;
	end process;

	-- data
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_d		<= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_d		<= (others => '0');
			elsif( (ram_we0 or ram_we1) = '1' )then
				ff_d		<= ram_d;
			else
				ff_d		<= (others => '0');
			end if;
		end if;
	end process;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_q		<= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_q		<= (others => '0');
			elsif( ff_oe_n = '0' )then
				ff_q		<= sram_dq;
			else
				ff_q		<= (others => '0');
			end if;
		end if;
	end process;

	-- ack
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_q_en		<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_q_en		<= '0';
			else
				ff_q_en		<= not ff_ce_n;
			end if;
		end if;
	end process;

	-- address
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_addr		<= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_addr		<= (others => '0');
			else
				ff_addr		<= ram_a;
			end if;
		end if;
	end process;
end rtl;
