-------------------------------------------------------------------------------
--	SimpleSystem Timer														 --
--	Copyright(C)2012 t.hara													 --
-------------------------------------------------------------------------------
--	History																	 --
--	2012/08/21	t.hara	first release										 --
-------------------------------------------------------------------------------
--	Address																	 --
--	00	rw	reset_count low 16bit											 --
--	01	rw	reset_count high 16bit											 --
--	10	rw	mode register													 --
--	11	w	start/stop														 --
-------------------------------------------------------------------------------
library	ieee;
	use	ieee.std_logic_1164.all;
	use	ieee.std_logic_unsigned.all;
	use	ieee.std_logic_arith.all;

entity simplesystem_timer is
	port(
		reset			: in	std_logic;
		clk				: in	std_logic;
		sreset			: in	std_logic;
		timer_a			: in	std_logic_vector(  1 downto 0 );
		timer_d			: in	std_logic_vector( 15 downto 0 );
		timer_oe		: in	std_logic;
		timer_we0		: in	std_logic;
		timer_we1		: in	std_logic;
		timer_q			: out	std_logic_vector( 15 downto 0 );
		timer_q_en		: out	std_logic;
		timer_irq		: out	std_logic
	);
end simplesystem_timer;

architecture rtl of simplesystem_timer is
	signal ff_irq			: std_logic;
	signal ff_irq_en		: std_logic;
	signal ff_active		: std_logic;
	signal ff_mode			: std_logic;		-- 0: one shot, 1: repeat
	signal ff_reset_count	: std_logic_vector( 31 downto 0 );
	signal ff_counter		: std_logic_vector( 31 downto 0 );
	signal ff_q				: std_logic_vector( 15 downto 0 );
	signal ff_q_en			: std_logic;
	signal w_q				: std_logic_vector( 15 downto 0 );
	signal w_all_zero		: std_logic;
	constant c_zero			: std_logic_vector( 31 downto 0 ) := (others => '0');
begin

	-- output assignment
	timer_irq	<= ff_irq and ff_irq_en;
	timer_q		<= ff_q;
	timer_q_en	<= ff_q_en;

	-- reset count
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_reset_count <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_reset_count <= (others => '0');
			elsif( timer_a = "00" and timer_we0 = '1' )then
				ff_reset_count(  7 downto  0 ) <= timer_d(  7 downto 0 );
			end if;
			if( timer_a = "00" and timer_we1 = '1' )then
				ff_reset_count( 15 downto  8 ) <= timer_d( 15 downto 8 );
			end if;
			if( timer_a = "01" and timer_we0 = '1' )then
				ff_reset_count( 23 downto 16 ) <= timer_d(  7 downto 0 );
			end if;
			if( timer_a = "01" and timer_we1 = '1' )then
				ff_reset_count( 31 downto 24 ) <= timer_d( 15 downto 8 );
			end if;
		end if;
	end process;

	-- mode
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_mode		<= '0';
			ff_irq_en	<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_mode		<= '0';
				ff_irq_en	<= '0';
			elsif( timer_a = "10" and timer_we0 = '1' )then
				ff_mode		<= timer_d(0);
				ff_irq_en	<= timer_d(1);
			end if;
		end if;
	end process;

	-- control
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_active	<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_active	<= '0';
			elsif( timer_a = "11" and timer_we0 = '1' )then
				if( timer_d(1) = '1' or timer_d(2) = '1' )then
					ff_active	<= (not timer_d(2)) and timer_d(1);
				end if;
			elsif( ff_mode = '0' and w_all_zero = '1' )then
					ff_active 	<= '0';
			end if;
		end if;
	end process;

	-- counter
	w_all_zero	<= '1' when( ff_counter = c_zero )else '0';

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_counter <= (others => '0');
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_counter <= (others => '0');
			elsif( timer_a = "11" and timer_we0 = '1' and timer_d(0) = '1' )then
				ff_counter <= ff_reset_count;
			elsif( ff_active = '1' )then
				if( w_all_zero = '1' )then
					if( ff_mode = '1' )then
						ff_counter <= ff_reset_count;
					end if;
				else
					ff_counter <= ff_counter - 1;
				end if;
			end if;
		end if;
	end process;

	-- interrupt
	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_irq	<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_irq	<= '0';
			elsif( ff_active = '1' and w_all_zero = '1' )then
				ff_irq	<= '1';
			elsif( timer_a = "11" and timer_we0 = '1' and  timer_d(3) = '1' )then
				ff_irq	<= '0';
			end if;
		end if;
	end process;

	-- read back
	with( timer_a )select w_q <=
		ff_counter( 15 downto  0 )						when "00",
		ff_counter( 31 downto 16 )						when "01",
		c_zero( 15 downto 2 ) & ff_irq_en & ff_mode		when "10",
		c_zero( 15 downto 1 ) & ff_irq					when "11",
		(others => '0')									when others;

	process( reset, clk )
	begin
		if( reset = '1' )then
			ff_q	<= (others => '0');
			ff_q_en	<= '0';
		elsif( clk'event and clk = '1' )then
			if( sreset = '1' )then
				ff_q	<= (others => '0');
				ff_q_en	<= '0';
			elsif( (timer_oe or timer_we0 or timer_we1) = '1' )then
				ff_q	<= w_q;
				ff_q_en	<= '1';
			else
				ff_q	<= (others => '0');
				ff_q_en	<= '0';
			end if;
		end if;
	end process;
end rtl;
